% AFMMGPU0 Computational experiment for the paper "Adaptive fast
% multipole method on the GPU".

% This code tests different values of ndirect to find the optimal value

if exist('./examples/data/AFMMgpu0.mat','file')
  warning('Existing file AFMMgpu0.mat NOT overwritten.');
end

clear all
rand('state',20120222);

%Parameters
Nmin=1e4;%minimum
Nmax=2e6;
Nstep=150;
tol=1e-6;
Nsets=5;
Naverages=10;
nn = round(logspace(log10(Nmin),log10(Nmax),Nstep));
ndir=5:75; %ndirect search interval

for k=1:length(nn)
  tic
  level=0;
  for m=1:Nsets
    vortices{k}=rand(1,nn(k))+1i*rand(1,nn(k));
    gammas{k}=rand(1,nn(k));
  end
  for h=1:length(ndir)
    %check if new value would give same number of levels. In that case,
    %copy results from last run
    newlevel=ceil(0.5*log2(5/8*nn(k)/ndir(h)));
    if newlevel==level && h ~=1
      cuda_ndirect(h,k,:,:,:)=cuda_ndirect(h-1,k,:,:,:);
      cpu_ndirect(h,k,:,:,:)=cpu_ndirect(h-1,k,:,:,:);
    else
      level=newlevel;
      for m=1:Nsets
        fmm2dCUDA(vortices{k},gammas{k},'pot',1,'ndirect',ndir(h),'tol',tol,'out','time');%dummy
        for n=1:Naverages
          cuda_ndirect(h,k,m,n,:)=fmm2dCUDA(vortices{k},gammas{k},'pot',1,'ndirect',ndir(h),'tol',tol,'out','time');
        end
        fmm2dCPU(vortices{k},gammas{k},'pot',1,'ndirect',ndir(h),'tol',tol,'out','time');%dummy
        for n=1:Naverages
          cpu_ndirect(h,k,m,n,:)=fmm2dCPU(vortices{k},gammas{k},'pot',1,'ndirect',ndir(h),'tol',tol,'out','time');
        end
      end
    end
  end
  disp(['step ' num2str(k) ' points ' num2str(nn(k)) ' time ' num2str(toc)]);
end

if ~exist('./examples/data/AFMMgpu0.mat','file')
  save examples/data/AFMMgpu0 cuda_ndirect cpu_ndirect nn ndir
end
