% AFMMGPU1 Computational experiment for the paper "Adaptive fast
% multipole method on the GPU".

% Break even plot
% This code times the code as a function of number of particles (N).
% Both direct evaluation and fmm evaluation is timed as a function of N.
% both on GPU and CPU

if exist('./examples/data/AFMMgpu1.mat','file')
  warning('Existing file AFMMgpu1.mat NOT overwritten.');
end

clear all
rand('state',20120223);

%Paramters
Nmin=1e3;       %minimum number of particles
Nmaxdirect=4e5; %direct evaluation only up to this point
Nmaxfmm=1e7;    %fmm evaluation up to this point
Nstep=70;       %steps in direct evaluation part
Nstep2=25;      %additional steps for fmm
tol=1e-6;
ndirectcpu=35;
ndirectgpu=45;
Nsets=5;        %averaging data
Naverages=10;
maxtime=50;     %maximum time for individual run
nn1 = round(logspace(log10(Nmin),log10(Nmaxdirect),Nstep));
nn2 = round(logspace(log10(Nmaxdirect),log10(Nmaxfmm),Nstep2));
nn2=[nn1 nn2(2:end)];
cuda_direct=zeros([length(nn1) Nsets Naverages 31]);
cuda_fmm=zeros([length(nn2) Nsets Naverages 31]);
cpu_fmm=zeros([length(nn2) Nsets Naverages 31]);
cpu_direct=zeros([length(nn1) Nsets Naverages 31]);

for k=1:length(nn1)
  tic
  for m=1:Nsets
    vortices=rand(1,nn2(k))+1i*rand(1,nn2(k));
    gammas=rand(1,nn2(k));
    if ~(size(cuda_direct,1)>=k && isnan(cuda_direct(k,1,1,1)))
      fmm2dCUDA(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      for n=1:Naverages
        cuda_direct(k,m,n,:)=fmm2dCUDA(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',0,'out','time');
      end
    end
    if ~(size(cuda_fmm,1)>=k && isnan(cuda_fmm(k,1,1,1)))
      fmm2dCUDA(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      for n=1:Naverages
        cuda_fmm(k,m,n,:)=fmm2dCUDA(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      end
    end
    if ~(size(cpu_fmm,1)>=k && isnan(cpu_fmm(k,1,1,1)))
      fmm2dCPU(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      for n=1:Naverages
        cpu_fmm(k,m,n,:)=fmm2dCPU(vortices,gammas,'pot',1,'ndirect',ndirectcpu,'tol',tol,'out','time');
      end
    end
    if ~(size(cpu_direct,1)>=k && isnan(cpu_direct(k,1,1,1)))
      fmm2dCPU(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      for n=1:Naverages
        cpu_direct(k,m,n,:)=fmm2dCPU(vortices,gammas,'pot',1,'ndirect',ndirectcpu,'tol',0,'out','time');
      end
    end
  end
  disp(['step ' num2str(k) ' points ' num2str(nn1(k)) ' time ' num2str(toc)]);
  if(mean(mean(cuda_direct(k,:,:,2)))>maxtime)
    for kk=k:length(nn1)
      cuda_direct(kk,:,:,:)=NaN;
    end
  end
  if(mean(mean(cuda_fmm(k,:,:,10)))>maxtime)
    for kk=k:length(nn2)
      cuda_fmm(kk,:,:,:)=NaN;
    end
  end
  if(mean(mean(cpu_fmm(k,:,:,10)))>maxtime)
    for kk=k:length(nn2)
      cpu_fmm(kk,:,:,:)=NaN;
    end
  end
  if(mean(mean(cpu_direct(k,:,:,2)))>maxtime)
    for kk=k:length(nn1)
      cpu_direct(kk,:,:,:)=NaN;
    end
  end
end

for k=length(nn1)+1:length(nn2)
  tic
  for m=1:Nsets
    vortices=rand(1,nn2(k))+1i*rand(1,nn2(k));
    gammas=rand(1,nn2(k));
    if ~(size(cuda_fmm,1)>=k && isnan(cuda_fmm(k,m,1,1)))
      fmm2dCUDA(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      for n=1:Naverages
        cuda_fmm(k,m,n,:)=fmm2dCUDA(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      end
    end
    if ~(size(cpu_fmm,1)>=k && isnan(cpu_fmm(k,m,1,1)))
      fmm2dCPU(vortices,gammas,'pot',1,'ndirect',ndirectgpu,'tol',tol,'out','time');
      for n=1:Naverages
        cpu_fmm(k,m,n,:)=fmm2dCPU(vortices,gammas,'pot',1,'ndirect',ndirectcpu,'tol',tol,'out','time');
      end
    end
  end
  disp(['step ' num2str(k) ' points ' num2str(nn2(k)) ' time ' num2str(toc)]);
  if(mean(mean(cuda_fmm(k,:,:,10)))>maxtime)
    for kk=k:length(nn2)
      cuda_fmm(kk,:,:,:)=NaN;
    end
  end
  if(mean(mean(cpu_fmm(k,:,:,10)))>maxtime)
    for kk=k:length(nn2)
      cpu_fmm(kk,:,:,:)=NaN;
    end
  end
end

if ~exist('./examples/data/AFMMgpu1.mat','file')
  save examples/data/AFMMgpu1 cuda_fmm cuda_direct cpu_fmm cpu_direct nn1 nn2
end
