function AFMMgpu1plots(errorplot)
% AFMMGPU1PLOTS Plots for the paper "Adaptive fast multipole method on
% the GPU".

%Break even plot
plotparams={'LineWidth',0.75};
load examples/data/AFMMgpu1
line1='-k';
line2='--b';
line3='-.k';
line4='.-b';
if nargin<1
  errorplot=false;
end
%calculate averages for each number of points
% cuda_fmm2=mean(cuda_fmm,3);
% cuda_direct2=mean(cuda_direct,3);
% cpu_fmm2=mean(cpu_fmm,3);
% cpu_direct2=mean(cpu_direct,3);
% cuda_fmm3=mean(cuda_fmm2,2);
% cuda_direct3=mean(cuda_direct2,2);
% cpu_fmm3=mean(cpu_fmm2,2);
% cpu_direct3=mean(cpu_direct2,2);
[cuda_fmm3 cuda_fmm3stdev]=AFMMgpu_post_process(cuda_fmm,1,1);
[cuda_direct3 cuda_direct3stdev]=AFMMgpu_post_process(cuda_direct,1,1);
[cpu_fmm3 cpu_fmm3stdev]=AFMMgpu_post_process(cpu_fmm,1,1);
[cpu_direct3 cpu_direct3stdev]=AFMMgpu_post_process(cpu_direct,1,1);

% production quality figure:
figure
loglog(nn2,cuda_fmm3(:,10),line1,nn2,cpu_fmm3(:,10),line2, ...
       nn1,cuda_direct3(:,2),line3,nn1,cpu_direct3(:,2),line4,plotparams{:});
h = legend('GPU FMM','CPU FMM','GPU Direct','CPU Direct', ...
           'Location','SE');
set(h,'FontSize',10);
if errorplot
  hold on;
  eline1 = [line1(end) '.']; % plot only bars and no graphs
  eline2 = [line2(end) '.'];
  eline3 = [line3(end) '.'];
  eline4 = [line4(end) '.'];

  nw = 20; % number of errorbars
  w1 = floor(numel(nn1)/nw);
  w2 = floor(numel(nn2)/nw);

  h1 = errorbar(nn2(1:w2:end),cuda_fmm3(1:w2:end,10), ...
                0.5*cuda_fmm3stdev(1:w2:end,10),eline1);
  h2 = errorbar(nn2(1:w2:end),cpu_fmm3(1:w2:end,10), ...
                0.5*cpu_fmm3stdev(1:w2:end,10),eline2);
  h3 = errorbar(nn1(1:w1:end),cuda_direct3(1:w1:end,2), ...
                0.5*cuda_direct3stdev(1:w1:end,2),eline3);
  % remove NaN's (cannot run CPU direct for all nn1's)
  iend = find(isnan(cpu_direct),1,'first');
  h4 = errorbar(nn1(1:w1:iend),cpu_direct3(1:w1:iend,2), ...
                0.5*cpu_direct3stdev(1:w1:iend,2),eline4);
  
  errorbar_tick(h1,0.25*nn2(1:w2:end),'units');
  errorbar_tick(h2,0.25*nn2(1:w2:end),'units');
  errorbar_tick(h3,0.25*nn1(1:w1:end),'units');
  errorbar_tick(h4,0.25*nn1(1:w1:iend),'units');
end

hold on,
loglog(nn1([5 40]),0.5e-8*nn1([5 40]).^2,'k','LineWidth',2)
h = text(2e3,0.2,'\propto N^2');
set(h,'FontSize',10);

loglog(nn2([50 90]),1e-7*nn2([50 90]),'k','LineWidth',2)
h = text(2e6,0.1,'\propto N');
set(h,'FontSize',10);

h = xlabel('N'); set(h,'FontSize',10);
h = ylabel('Time (s)'); set(h,'FontSize',10);
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 250]);
axis([1e3 1e7 1e-4 1e1]);
set(gca,'ytick',10.^[-4:1]);

print -depsc examples/fig/breakevenplot.eps

figure
semilogx(nn1,cpu_direct3(:,2)./cuda_direct3(:,2),line1, ...
         nn2,cpu_fmm3(:,10)./cuda_fmm3(:,10),line2,plotparams{:});
if errorplot
  hold on;
  
  [direct3 direct3stdev] = getdivstdev(cpu_direct3(:,2), ...
                                       cuda_direct3(:,2), ...
                                       cpu_direct3stdev(:,2), ...
                                       cuda_direct3stdev(:,2));
  h1 = errorbar(nn1(1:w1:iend),direct3(1:w1:iend), ...
                0.5*direct3stdev(1:w1:iend),eline1);

  [fmm3 fmm3stdev] = getdivstdev(cpu_fmm3(:,10), ...
                                 cuda_fmm3(:,10), ...
                                 cpu_fmm3stdev(:,10), ...
                                 cuda_fmm3stdev(:,10));
  h2 = errorbar(nn2(1:w2:end),fmm3(1:w2:end), ...
                0.5*fmm3stdev(1:w2:end),eline2);

  errorbar_tick(h1,0.25*nn1(1:w1:iend),'units');
  errorbar_tick(h2,0.25*nn2(1:w2:end),'units');
end
h = xlabel('N');
set(h,'FontSize',10);
h = ylabel('Speedup'); set(h,'FontSize',10);
h = legend('Direct','FMM'); set(h,'FontSize',10);
axis([1e3 1e7 0 20]);
set(gcf,'PaperPositionMode','auto');
set(gcf,'Position',[100 100 380 200]);
print -depsc examples/fig/breakevenspeedup.eps
