% AFMMGPU5 Computational experiment for the paper "Adaptive fast
% multipole method on the GPU".

% Timing as a function of pcoeff
% this function determines the timing of the individual parts of the
% algoritym as a function of the number of multipole coefficients
% (tolerance)
if exist('./examples/data/AFMMgpu5.mat','file')
  warning('Existing file AFMMgpu5.mat NOT overwritten.');
end

clear all
rand('state',20120227);

%Parameters
pmin=2; %lowest value for number of coefficients
pmax=52;%highers value
theta=0.5;%theta value according to make file
pp=pmin:pmax;
tols=2.^((pp+0.5).*log2(theta)-2);
ndirect=45;
Nsets=5;
Naverages=10;
Nvort=1e6;
vdummy=rand(1,1e3)+1i*rand(1,1e3);
gdummy=rand(1,1e3);

for m=1:Nsets
  vortices{m}=rand(1,Nvort)+1i*rand(1,Nvort);
  gammas{m}=rand(1,Nvort);
end

for k=1:length(tols)
  tic
  for m=1:Nsets
    fmm2dCUDAnosort( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cuda_nosort_timing_tol(k,m,n,:)=fmm2dCUDAnosort( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time');
    end
    fmm2dCUDA( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cuda_timing_tol(k,m,n,:)=fmm2dCUDA( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time');
    end
    fmm2dCPUnompeval( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cpu_nompeval_timing_tol(k,m,n,:)=fmm2dCPUnompeval( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time');
    end
    fmm2dCPUnodirect( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time'); %one call before timing to access all elements in array first
    for n=1:Naverages
      cpu_nodirect_timing_tol(k,m,n,:)=fmm2dCPUnodirect( vortices{m},gammas{m},'pot',1,'ndirect',ndirect,'tol',tols(k),'out','time');
    end
  end
  disp(['step ' num2str(k) ' tol ' num2str(tols(k)) ' time ' num2str(toc)]);
end

if ~exist('./examples/data/AFMMgpu5.mat','file')
  save examples/data/AFMMgpu5 cuda_timing_tol cuda_nosort_timing_tol cpu_nompeval_timing_tol cpu_nodirect_timing_tol tols pp
end
