% evaltimefarfielddirect_run Runs for the paper "A General high order two-
% dimensional panel method".
% 
% This function is used to measure the time it takes for evaluating the
% contributions of panels in the far field using direct evaluation. 
%
% The parameters are
% 
% Nrep - Number of repetitions for each time measurement
% Neval - Number of evaluation positions
% vortexpanels - if the panels are made from vorticity (1) or source
%                strength(2)
% controlpoint - If control points in the center of the panel, or total
%                flow throught the panel will be used
% slopecorrection - If the strength of the panel should be corrected
% depending on the panel slope
% doublecontrol - Adds extra target panel and solve equations in the least
%                 square sense
function evaltimefarfielddirect_run(Nrep,Neval,vortexpanels,controlpoint,slopecorrection,doublecontrol)
paneleval=10:500;

% Set up panels separately, this will not be included in the evaluation
% time
for m=1:length(paneleval)
  % Set up parameters for building the circle. Note that higher order
  % derivates are these two with negative sign
  anglesd=linspace(0,2*pi,paneleval(m)+1);
  panelpointsd=exp(1i*anglesd(1:end-1));
  paneldirectionsd=1i*exp(1i*anglesd(1:end-1));
  linpanels{m}=highorderpanelstrength(panelpointsd,[],[],[],1,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  cubpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,[],[],1,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  quintpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,-panelpointsd,[],1,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  septpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,-panelpointsd,-paneldirectionsd,1,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  lincubpanels{m}=highorderpanelstrength(panelpointsd,[],[],[],3,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  cubcubpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,[],[],3,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  quintcubpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,-panelpointsd,[],3,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  septcubpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,-panelpointsd,-paneldirectionsd,3,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  linquintpanels{m}=highorderpanelstrength(panelpointsd,[],[],[],5,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  cubquintpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,[],[],5,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  quintquintpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,-panelpointsd,[],5,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
  septquintpanels{m}=highorderpanelstrength(panelpointsd,paneldirectionsd,-panelpointsd,-paneldirectionsd,5,1,vortexpanels,controlpoint,slopecorrection,doublecontrol);
end

% Run all tests and measure the time
for k=1:Nrep
  eval=(9*rand(1,Neval)+1).*exp(2i*pi*rand(1,Neval));
  for m=1:length(paneleval)
    
    lintimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'linshape',linpanels{m}.panels,'linstrength',linpanels{m}.strength});
    cubtimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'cubshape',cubpanels{m}.panels,cubpanels{m}.directions,'linstrength',cubpanels{m}.strength});
    quinttimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'quintshape',quintpanels{m}.panels,quintpanels{m}.directions,quintpanels{m}.directions2,'linstrength',quintpanels{m}.strength});
    septtimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'septshape',septpanels{m}.panels,septpanels{m}.directions,septpanels{m}.directions2,septpanels{m}.directions3,'linstrength',septpanels{m}.strength});
    lincubtimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'linshape',lincubpanels{m}.panels,'cubstrength',lincubpanels{m}.strength,lincubpanels{m}.strengthdir});
    cubcubtimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'cubshape',cubcubpanels{m}.panels,cubcubpanels{m}.directions,'cubstrength',cubcubpanels{m}.strength,cubcubpanels{m}.strengthdir});
    quintcubtimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'quintshape',quintcubpanels{m}.panels,quintcubpanels{m}.directions,quintcubpanels{m}.directions2,'cubstrength',quintcubpanels{m}.strength,quintcubpanels{m}.strengthdir});
    septcubtimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'septshape',septcubpanels{m}.panels,septcubpanels{m}.directions,septcubpanels{m}.directions2,septcubpanels{m}.directions3,'cubstrength',septcubpanels{m}.strength,septcubpanels{m}.strengthdir});
    linquinttimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'linshape',linquintpanels{m}.panels,'quintstrength',linquintpanels{m}.strength,linquintpanels{m}.strengthdir,linquintpanels{m}.strength2dir});
    cubquinttimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'cubshape',cubquintpanels{m}.panels,cubquintpanels{m}.directions,'quintstrength',cubquintpanels{m}.strength,cubquintpanels{m}.strengthdir, cubquintpanels{m}.strength2dir});
    quintquinttimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'quintshape',quintquintpanels{m}.panels,quintquintpanels{m}.directions,quintquintpanels{m}.directions2,'quintstrength',quintquintpanels{m}.strength,quintquintpanels{m}.strengthdir, quintquintpanels{m}.strength2dir});
    septquinttimedirect(m,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'out','time','ndirect',35,'destinationmesh',true,'sourcepanels',{'septshape',septquintpanels{m}.panels,septquintpanels{m}.directions,septquintpanels{m}.directions2,septquintpanels{m}.directions3,'quintstrength',septquintpanels{m}.strength,septquintpanels{m}.strengthdir, septquintpanels{m}.strength2dir});
    

  end
end

if ~exist('./examples/generalpanel/data/generalpanelfarfielddirect.mat','file')
  save examples/generalpanel/data/generalpanelfarfielddirect lintimedirect cubtimedirect quinttimedirect septtimedirect lincubtimedirect cubcubtimedirect quintcubtimedirect septcubtimedirect linquinttimedirect cubquinttimedirect quintquinttimedirect septquinttimedirect Neval Nrep paneleval
end
