% evaltimenearfield_run Runs for the paper "A General high order two-
% dimensional panel method".
% 
% This function is used to measure the time it takes for evaluating the
% contributions of panels in the near field using direct evaluation. Panels
% are generated with random coefficients over the diagonal of the unitbox
% and all evaluation positions are inside the box
%
% The parameters are
% 
% shapeN - Maximum number of coefficients for the shape
% strN - Maximum number of coefficients for the strength
% Npanel - Number of panels
% Nrep - Number of repetitions
% 
% It is recommended to run with a relatively large number of panels, as the
% evaluation time depends on the shape of the panel (if it hits any of the
% numerical special cases)
function evaltimenearfield_run(shapeN,strN,Npanel,Neval,Nrep)
% Make repetitions in the outer loop to avoid issues if the computer
% temprary runs slower
for k=1:Nrep
  panels=[1;1i]*ones(1,Npanel);
  eval=rand(1,Neval)+1i*rand(1,Neval);
  panela{1}=zeros(2,Npanel);
  % loop ofer the shapes
  for m=1:shapeN
    % Build panels
    if m==1 % Linear panels have to be exactly the diagonal
      panela{1}=zeros(2,Npanel);
    else
      panela{m}=2*(rand(m,Npanel)-0.5);
      % Adapt last coefficients to ensure that the panel ends in the other
      % corner
      for n=1:Npanel
        panela{m}(m+1,n)=-polyval(flipud(panela{m}(1:m,n)),sqrt(2))/sqrt(2)^m;
      end
    end
    for n=0:strN
      % Random strength coefficients. These values should not affect the
      % simulation time
      panelb{n+1}=rand(n+1,Npanel);
      % Measure the time
      time(m,n+1,k,:)=fmm2d([],[],eval,'pot',1,'tol',0,'sourcepanels',{'polyshape',panels,panela{m},'polystrength',panelb{n+1}},'out','time');
    end
  end
end
if ~exist('./examples/generalpanel/data/generalpanelnearfield.mat','file')
  save examples/generalpanel/data/generalpanelnearfield time shapeN strN Npanel Neval Nrep
end
