% generalpanelcontrolpoint
%
% This function calculates the position and normal direction of control
% points on the panel surface
%
% The parameters are
% 
% points - where control points should be added, normalized values between
%          [0, 1]
% panels - the start and end points of the panels
% a      - polynomial for the panel shape

function [pt, n]=generalpanelcontrolpoint(points,panels,a)
theta=angle(panels(2,:)-panels(1,:));
deltas=abs(panels(2,:)-panels(1,:));

pt=zeros(size(panels,2)*length(points),1);
n=zeros(size(panels,2)*length(points),1);
ptsize=length(points);
if size(a,1)==1
  ap=zeros(1,size(a,2));
else
  ap=zeros(size(a,1)-1,size(a,2)); % calculate derivative (for normal vector)
  for k=2:size(a,1)
    ap(k-1,:)=(k-1)*a(k,:);
  end
end
for k=1:size(panels,2)
  af=flipud(a(:,k));  % polyval stores coefficients in the opposite order as the fmm code
  afd=flipud(ap(:,k));
  for m=1:ptsize
    localpoint=points(m)*deltas(k)+1i*polyval(af,points(m)*deltas(k));
    localderiv=1+1i*polyval(afd,points(m)*deltas(k));
    pt((k-1)*ptsize+m)=panels(1,k)+(localpoint)*exp(1i*theta(k));
    n((k-1)*ptsize+m)=-1i*localderiv*exp(1i*theta(k));
  end
end
n=n./abs(n);