% generalpaneltargetrhs
%
% Calculates the right hand side to use for solving the panel matrix, under
% the condition that the only flow contribution is the asymptotic flow
% velocity, and that the panels is standing still. In this case, it is
% sufficient to calculate the flow through a point in the center of the
% panel
%
% The parameters are
% 
% panels         - the start and end points of the panels
% a              - polynomial for the panel shape
% vasympt        - asymptotic flow velocity
% kuttacondition - how the kutta condition is solved (determines the number of unknown) 
function rhs=generaltargetpanelrhs(panels,a,vasympt,kuttacondition)
theta=angle(panels(2,:)-panels(1,:));
deltas=abs(panels(2,:)-panels(1,:));

rhs=zeros(size(panels,2)+1-kuttacondition,1);
for k=1:size(panels,2)
    posstart=panels(1,k)+a(1,k)*exp(1i*theta(k)); % the panels values are only describe the reference line, calculate the actual start and end positions
    posend=panels(1,k)+(deltas(k)+1i*polyval(flipud(a(:,k)),deltas(k)))*exp(1i*theta(k));
    n=-1i*(posend-posstart);  % Normal vector
    n=n/abs(n);
    rhs(k)=-real(vasympt.*abs(posend-posstart)*n)*2*pi; % Scalar product of flow velocity multiplied with panel length
end