% This is a function to calculate the inverse Karmat Trefftz mapping, going
% from z coordinates to s cordinates. The process is iterative.
%
% The parameters are
%
% z - z coordinates
% n - Coefficient for the Karman Trefftz mapping
% my - Center of the circle in the s-plane for the Karman Trefftz mapping
% startguess - Initial value for the itertive process

function s=inversekarmantrefftz(z,n,my,startguess)
s=zeros(size(z));
R=sqrt((1-real(my)).^2+imag(my).^2);
msum=0;
for k=1:numel(z) % Loop over all elements
  m=1;
  snew=startguess(k);
  s0=snew;
  reset=true;
  
  while m<500 && (abs((snew-s0)/s0)>1e-14 || reset)
    svalue=karmantrefftz(snew,n)-z(k); % The error in z coordinate
    if (m<11 || reset) && abs(snew-my)<R % Value inside the circle, mirror it
      snew=R/(snew-my)+my;
      s0=snew;
      svalue=karmantrefftz(snew,n)-z(k);
    end
    if (m>10 && ~reset) && (abs(svalueold)<abs(svalue) || abs(snew-my)<R) % Apply damping if solution is not converging
      sstep=0.5*sstep;
    else % Calculate the step to take
      svalueold=svalue;
      s0=snew;
      sstep=svalue/karmantrefftzd(s0,n); % Correct the error by the derivative of the mapping
    end
    snew=s0-sstep; % new value for the interation
    reset=false;
    m=m+1;
    if m==150      % If it hasn't converged yet, restart it with new guess
      snew=z(k);
      if abs(snew-my)<R
        snew=R/(snew-my)+my;
      end
      s0=snew;
      reset=true;
    end
    if m==200 || ((abs((snew-s0)/s0)<1e-13 && abs(svalue)>1e-7) && ~reset) % Another reset condition in case of covergence issues, here if snew goes toward s0, without any improvement in the error
      snew=startguess(k);
      if abs(snew-my)<R
        snew=R/(snew-my)+my;
      end
      s0=snew;
      reset=true;
      p=1;
      while p<500 && abs((snew-s0)/s0)>1e-14 || reset
        svalue=karmantrefftz(snew,n)-z(k);
        if p>1 && (abs(svalueold)<abs(svalue) || abs(snew-my)<R) %here, run with possible damping from the start
          sstep=0.5*sstep;
        else
          svalueold=svalue;
          s0=snew;
          sstep=svalue;
        end
        snew=s0-sstep;
        reset=false;
        p=p+1;
      end
      reset=true;
    end
    if m==100 % yet another reset condition
      q=1;
      snew=startguess(k);
      s0=snew;
      reset=true;
      while q<500 && abs((snew-s0)/s0)>1e-14 || reset
        svalue=karmantrefftz(snew,n)-z(k);
        if q>10 && ~reset && abs(svalueold)<abs(svalue)
          sstep=0.5*sstep;
        else
          svalueold=svalue;
          s0=snew;
          sstep=svalue/karmantrefftzd(s0,n);
        end
        snew=s0-sstep;
        reset=false;
        q=q+1;
      end
      if abs(snew-my)<R % If the new value is within the circle
        pole=snew;
        q=1;
        snew=startguess(k);
        s0=snew;
        reset=true;
        while q<500 && abs((snew-s0)/s0)>1e-14 || reset
          svalue=(karmantrefftz(snew,n)-z(k))/(snew-pole);
          if q>10 && ~reset && abs(svalueold)<abs(svalue)
            sstep=0.5*sstep;
          else
            svalueold=svalue;
            s0=snew;
            sstep=svalue/(((s0-pole)*karmantrefftzd(s0,n)-karmantrefftz(s0,n)+z(k))/(s0-pole)^2);
          end
          snew=s0-sstep;
          reset=false;
          q=q+1;
        end
      end
    end
  end
  s(k)=snew;
  msum=msum+m;
end