% karmantrefftzconvergence_plots Plots for the paper "A General high order 
% two-dimensional panel method".
%
% Plots for the Karman Trefftz tests.
% This requires that karmantrefftzconvergence_run has been called before.

function karmantrefftzconvergence_plot(filename)
load(filename)

pos=[94   391   650   504];
legendpos=[0.70974358974359 0.478174603174603 0.255384615384615 0.498677248677248];

if diststudy || surfacestudy
  for k=1:length(panelcount)
    % How the error varies along the integration line
    if surfacestudy
      figure
      semilogy(surface_pos{k},linerror_surf{k},'b','LineWidth',1);
      hold on;
      semilogy(surface_pos{k},cuberror_surf{k},'-.b','LineWidth',1);
      semilogy(surface_pos{k},quinterror_surf{k},'--b','LineWidth',1);
      semilogy(surface_pos{k},septerror_surf{k},':b','LineWidth',1);
      semilogy(surface_pos{k},lincuberror_surf{k},'r','LineWidth',1);
      semilogy(surface_pos{k},cubcuberror_surf{k},'-.r','LineWidth',1);
      semilogy(surface_pos{k},quintcuberror_surf{k},'--r','LineWidth',1);
      semilogy(surface_pos{k},septcuberror_surf{k},':r','LineWidth',1);
      semilogy(surface_pos{k},linquinterror_surf{k},'g','LineWidth',1);
      semilogy(surface_pos{k},cubquinterror_surf{k},'-.g','LineWidth',1);
      semilogy(surface_pos{k},quintquinterror_surf{k},'--g','LineWidth',1);
      semilogy(surface_pos{k},septquinterror_surf{k},':g','LineWidth',1);
      semilogy(surface_pos{k},linsepterror_surf{k},'m','LineWidth',1);
      semilogy(surface_pos{k},cubsepterror_surf{k},'-.m','LineWidth',1);
      semilogy(surface_pos{k},quintsepterror_surf{k},'--m','LineWidth',1);
      semilogy(surface_pos{k},septsepterror_surf{k},':m','LineWidth',1);

      title(['Distance: ' num2str(surfaceval) ' controlpoints: ' num2str(controlpoint) ' slopecorr: ' num2str(slopecorrection)]);
      legend('M=1, N=1','M=3, N=1','M=5, N=1','M=7, N=1', ...
      'M=1, N=3','M=3, N=3','M=5, N=3','M=7, N=3',...
      'M=1, N=5','M=3, N=5','M=5, N=5','M=7, N=5',...
      'M=1, N=7','M=3, N=7','M=5, N=7','M=7, N=7');
      legend1 = legend(gca,'show');
      set(legend1,'Position',legendpos);
      set(gcf,'Position',pos);
      set(gca,'FontSize',14);
      xlabel('Position along airfoil','FontSize',14);
      ylabel('Error','FontSize',14);
    end

    % How the error varies as a function of distance from the airfoil
    if diststudy
      figure
      loglog(sp,linerror{k},'b','LineWidth',1);
      hold on;
      loglog(sp,cuberror{k},'-.b','LineWidth',1);
      loglog(sp,quinterror{k},'--b','LineWidth',1);
      loglog(sp,septerror{k},':b','LineWidth',1);
      loglog(sp,lincuberror{k},'r','LineWidth',1);
      loglog(sp,cubcuberror{k},'-.r','LineWidth',1);
      loglog(sp,quintcuberror{k},'--r','LineWidth',1);
      loglog(sp,septcuberror{k},':r','LineWidth',1);
      loglog(sp,linquinterror{k},'g','LineWidth',1);
      loglog(sp,cubquinterror{k},'-.g','LineWidth',1);
      loglog(sp,quintquinterror{k},'--g','LineWidth',1);
      loglog(sp,septquinterror{k},':g','LineWidth',1);
      loglog(sp,linsepterror{k},'m','LineWidth',1);
      loglog(sp,cubsepterror{k},'-.m','LineWidth',1);
      loglog(sp,quintsepterror{k},'--m','LineWidth',1);
      loglog(sp,septsepterror{k},':m','LineWidth',1);
      title(['Distance: ' num2str(distval(k)) ' controlpoints: ' num2str(controlpoint) ' slopecorr: ' num2str(slopecorrection)]);
      legend('M=1, N=1','M=3, N=1','M=5, N=1','M=7, N=1', ...
      'M=1, N=3','M=3, N=3','M=5, N=3','M=7, N=3',...
      'M=1, N=5','M=3, N=5','M=5, N=5','M=7, N=5',...
      'M=1, N=7','M=3, N=7','M=5, N=7','M=7, N=7');
      legend1 = legend(gca,'show');
      set(legend1,'Position',legendpos);
      set(gcf,'Position',pos);
      set(gca,'FontSize',14);
      xlabel('Distance','FontSize',14);
      ylabel('Error','FontSize',14);
    end
  end
end

% Accuracy as function of number of panels
if panelstudy
  for k=1:length(distval)
    % Accuracy in flow velocity
    if velstudy
      figure
      loglog(paneleval,linerrord{k},'b','LineWidth',1);
      hold on;
      loglog(paneleval,cuberrord{k},'-.b','LineWidth',1);
      loglog(paneleval,quinterrord{k},'--b','LineWidth',1);
      loglog(paneleval,septerrord{k},':b','LineWidth',1);
      loglog(paneleval,lincuberrord{k},'r','LineWidth',1);
      loglog(paneleval,cubcuberrord{k},'-.r','LineWidth',1);
      loglog(paneleval,quintcuberrord{k},'--r','LineWidth',1);
      loglog(paneleval,septcuberrord{k},':r','LineWidth',1);
      loglog(paneleval,linquinterrord{k},'g','LineWidth',1);
      loglog(paneleval,cubquinterrord{k},'-.g','LineWidth',1);
      loglog(paneleval,quintquinterrord{k},'--g','LineWidth',1);
      loglog(paneleval,septquinterrord{k},':g','LineWidth',1);
      loglog(paneleval,linsepterrord{k},'m','LineWidth',1);
      loglog(paneleval,cubsepterrord{k},'-.m','LineWidth',1);
      loglog(paneleval,quintsepterrord{k},'--m','LineWidth',1);
      loglog(paneleval,septsepterrord{k},':m','LineWidth',1);
      xlim([10 300]);
      
      title(['Distance: ' num2str(distval(k)) ' controlpoints: ' num2str(controlpoint) ' slopecorr: ' num2str(slopecorrection)]);
      legend('M=1, N=1','M=3, N=1','M=5, N=1','M=7, N=1', ...
        'M=1, N=3','M=3, N=3','M=5, N=3','M=7, N=3',...
        'M=1, N=5','M=3, N=5','M=5, N=5','M=7, N=5',...
        'M=1, N=7','M=3, N=7','M=5, N=7','M=7, N=7');
      legend1 = legend(gca,'show');
      set(legend1,'Position',legendpos);
      set(gcf,'Position',pos);
      set(gca,'FontSize',14);
      xlabel('Number of panels','FontSize',14)
      ylabel('Error','FontSize',14)
    end
  end
  % Accuracy in circulation
  if circstudy
    figure
    loglog(paneleval,linmyerror,'b','LineWidth',1);
    hold on;
    loglog(paneleval,cubmyerror,'-.b','LineWidth',1);
    loglog(paneleval,quintmyerror,'--b','LineWidth',1);
    loglog(paneleval,septmyerror,':b','LineWidth',1);
    loglog(paneleval,lincubmyerror,'r','LineWidth',1);
    loglog(paneleval,cubcubmyerror,'-.r','LineWidth',1);
    loglog(paneleval,quintcubmyerror,'--r','LineWidth',1);
    loglog(paneleval,septcubmyerror,':r','LineWidth',1);
    loglog(paneleval,linquintmyerror,'g','LineWidth',1);
    loglog(paneleval,cubquintmyerror,'-.g','LineWidth',1);
    loglog(paneleval,quintquintmyerror,'--g','LineWidth',1);
    loglog(paneleval,septquintmyerror,':g','LineWidth',1);
    loglog(paneleval,linseptmyerror,'m','LineWidth',1);
    loglog(paneleval,cubseptmyerror,'-.m','LineWidth',1);
    loglog(paneleval,quintseptmyerror,'--m','LineWidth',1);
    loglog(paneleval,septseptmyerror,':m','LineWidth',1);
    xlim([10 300]);
    
    title(['Circulation: controlpoints: ' num2str(controlpoint) ' slopecorr: ' num2str(slopecorrection)]);
    legend('M=1, N=1','M=3, N=1','M=5, N=1','M=7, N=1', ...
      'M=1, N=3','M=3, N=3','M=5, N=3','M=7, N=3',...
      'M=1, N=5','M=3, N=5','M=5, N=5','M=7, N=5',...
      'M=1, N=7','M=3, N=7','M=5, N=7','M=7, N=7');
    legend1 = legend(gca,'show');
    set(legend1,'Position',legendpos);
    set(gcf,'Position',pos);
    set(gca,'FontSize',14);
  end
end