% karmantrefftzconvergenccpsce_plots Plots for the paper "A General high order two-
% dimensional panel method".
%
% Plots for the comparing the different panel implementations
% This requires that karmantrefftzconvergencecpsc_run has been called before.
function karmantrefftzconvergencecpsc_plot(filename)
load(filename);

pos=[94   391   650   504];
legendpos=[0.70974358974359 0.478174603174603 0.255384615384615 0.498677248677248];

if diststudy || surfacestudy
  for k=1:length(panelcount)
    % How the error varies along the integration line
    if surfacestudy

      figure
      semilogy(surface_pos{k},cubcuberror_surf{k},'-k','LineWidth',1);
      hold on;
      semilogy(surface_pos{k},quintquinterror_surf{k},'--k','LineWidth',1);
      semilogy(surface_pos{k},cubcuberrorcp_surf{k},'-b','LineWidth',1);
      semilogy(surface_pos{k},quintquinterrorcp_surf{k},'--b','LineWidth',1);
      semilogy(surface_pos{k},cubcuberrorsc_surf{k},'-r','LineWidth',1);
      semilogy(surface_pos{k},quintquinterrorsc_surf{k},'--r','LineWidth',1);
      title(['Distance: ' num2str(surfaceval)]);
      legend('M=3, N=3','M=5, N=5','M=3, N=3 cp','M=5, N=5 cp', ...
        'M=3, N=3 sc','M=5, N=5 sc');
      legend1 = legend(gca,'show');
      set(legend1,'Position',legendpos);
      set(gcf,'Position',pos);
      set(gca,'FontSize',14);
      xlabel('Position along airfoil','FontSize',14);
      ylabel('Error','FontSize',14);
    end

    % How the error varies as a function of distance from the airfoil
    if diststudy
      figure
      loglog(sp(2:end),cubcuberror{k}(2:end),'-k','LineWidth',1);
      hold on;
      loglog(sp,quintquinterror{k},'--k','LineWidth',1);
      loglog(sp(2:end),cubcuberrorcp{k}(2:end),'-b','LineWidth',1);
      loglog(sp,quintquinterrorcp{k},'--b','LineWidth',1);
      loglog(sp(2:end),cubcuberrorsc{k}(2:end),'-r','LineWidth',1);
      loglog(sp,quintquinterrorsc{k},'--r','LineWidth',1);
      
      
      title(['Distance: ' num2str(distval(k))]);
      legend('M=3, N=3','M=5, N=5','M=3, N=3 cp','M=5, N=5 cp', ...
        'M=3, N=3 sc','M=5, N=5 sc');
      legend1 = legend(gca,'show');
      set(legend1,'Position',legendpos);
      set(gcf,'Position',pos);
      set(gca,'FontSize',14);
      xlabel('Distance','FontSize',14);
      ylabel('Error','FontSize',14);
    end
  end
end

% Accuracy as function of number of panels
if panelstudy
  for k=1:length(distval)
    % Accuracy in flow velocity
    if velstudy
      figure
      loglog(paneleval,cubcuberrord{k},'-k','LineWidth',1);
      hold on;
      loglog(paneleval,quintquinterrord{k},'--k','LineWidth',1);
      loglog(paneleval,cubcuberrordcp{k},'-b','LineWidth',1);
      loglog(paneleval,quintquinterrordcp{k},'--b','LineWidth',1);
      loglog(paneleval,cubcuberrordsc{k},'-r','LineWidth',1);
      loglog(paneleval,quintquinterrordsc{k},'--r','LineWidth',1);
      xlim([10 300]);
      
      title(['Distance: ' num2str(distval(k))]);
      legend('M=3, N=3','M=5, N=5','M=3, N=3 cp','M=5, N=5 cp', ...
        'M=3, N=3 sc','M=5, N=5 sc');
      legend1 = legend(gca,'show');
      set(legend1,'Position',legendpos);
      set(gcf,'Position',pos);
      set(gca,'FontSize',14);
      xlabel('Number of panels','FontSize',14)
      ylabel('Error','FontSize',14)
    end
  end
  
  % Accuracy in circulation
  if circstudy
    figure
    loglog(paneleval,cubcubmyerror{k},'-k','LineWidth',1);
    hold on;
    loglog(paneleval,quintquintmyerror{k},'--k','LineWidth',1);
    loglog(paneleval,cubcubmyerrorcp{k},'-b','LineWidth',1);
    loglog(paneleval,quintquintmyerrorcp{k},'--b','LineWidth',1);
    loglog(paneleval,cubcubmyerrorsc{k},'-r','LineWidth',1);
    loglog(paneleval,quintquintmyerrorsc{k},'--r','LineWidth',1);
    xlim([10 300]);
    
    title('Circulation');
    legend('M=3, N=3','M=5, N=5','M=3, N=3 cp','M=5, N=5 cp', ...
        'M=3, N=3 sc','M=5, N=5 sc');
  end
end