% lineartargetpanelrhscontrolpoint
%
% Calculates the right hand side to use for solving the panel matrix, under
% the condition that the only flow contribution is the asymptotic flow
% velocity, and that the panels is standing still. In this case, it is
% sufficient to calculate the flow through a point in the center of the
% panel
%
% The parameters are
% 
% controlpoints  - positions where the flow velocity is evaluated
% n              - normal vectors
% vasympt        - asymptotic flow velocity
% kuttacondition - how the kutta condition is solved (determines the number of unknown) 

function rhs=lineartargetpanelrhscontrolpoint(controlpoints,n,vasympt,kuttacondition)
rhs=zeros(length(controlpoints)+1-kuttacondition,1);
rhs(1:end-1+kuttacondition)=-real(conj(vasympt).*n)*2*pi;
