/* mexpanel.cpp */

/* S. Engblom and A. Goude 2011-06-15 */

#include <string.h>
#include "matrix.h"
#include "mex.h"
#include "generalpanel.h"
#include "mexgeneralpanel.h"

// macros to be independent of fmm2d.cpp
#define ISDOUBLEMATRIX(A) (mxIsDouble(A) && !mxIsSparse(A) && \
                           mxGetNumberOfDimensions(A) == 2)
#define ISREALMATRIX(A) (ISDOUBLEMATRIX(A) && !mxIsComplex(A))
#define ISREALSCALAR(A) (ISDOUBLEMATRIX(A) && !mxIsComplex(A) && mxGetNumberOfElements(A)==1)
#define ISLOGICALMATRIX(A) (mxIsLogical(A) && !mxIsSparse(A) && \
                            mxGetNumberOfDimensions(A) == 2)
#define ISLOGICALSCALAR(A) (mxIsLogical(A) && !mxIsSparse(A) && mxGetNumberOfElements(A)==1)

typedef enum {QUINTSTRENGTH = 0,QUINTSHAPE = 3,LINSTRENGTH = 4,NORMALIZED = 5,SHIFTED = 6,
              OUTSIDE = 7,CUBSHAPE = 8,POLYSHAPE = 9,POLYSTRENGTH = 10,SEPTSHAPE = 12,
              SEPTSTRENGTH = 13,CUBSTRENGTH = 16,LINSHAPE = 17,SLOPECORRECTION = 19, ERR_PROPERTY = -1} PANELPROPERTY;
/*------------------------------------------------------------------------*/
void getSOURCEPANEL(void *panels_,const mxArray *rhs)
{
  panellist* panels = (panellist *)panels_;
  double outside=0.0;
  panels->slopecorrection=0;
  panels->aN=0;
  panels->bN=0;

  static const getgeneralpanelvalFun getval[] = {&getQUINTSTRENGTH,NULL,NULL,&getQUINTSHAPE,&getLINSTRENGTH,&getNORMALIZED,&getSHIFTED,
                                                 &getOUTSIDE,&getCUBSHAPE,&getPOLYSHAPE,&getPOLYSTRENGTH,NULL,&getSEPTSHAPE,
                                                 &getSEPTSTRENGTH,NULL,NULL,&getCUBSTRENGTH,&getLINSHAPE,NULL,&getSLOPECORRECTION,NULL};
  void *val[] = {panels,NULL,NULL,panels,panels,panels,panels,
                 &outside,panels,panels,panels,NULL,panels,
                 panels,NULL,NULL,panels,panels,NULL,panels,NULL};
  if(!mxIsCell(rhs))
    mexErrMsgTxt("sourcepanel data must be given as cell object");
  int npanelrhs=mxGetNumberOfElements(rhs);
  // parse panel property/value-combinations
  for (int j = 0; j < npanelrhs-1; j++) {
    int prop;
    char str[17];
    const mxArray *Prop = mxGetCell(rhs,j);

    // determine property
    if (!mxIsChar(Prop))
      mexErrMsgTxt("Expecting a panel property name.");
    if (mxGetString(Prop,str,16) != 0 ||
      (prop = getgeneralpanelprop(str)) == ERR_PROPERTY)
      mexErrMsgTxt("Unknown panel property.");

    // get its value
    j+=getval[prop](val[prop],rhs,j+1,npanelrhs);
  }

  //finalize
  if(panels->aN==0)
    mexErrMsgTxt("Panel shape must be given");
  if(panels->bN==0)
    mexErrMsgTxt("Panel strength must be given");
  if(outside!=0.0) {
    for(int i=0;i<panels->count;i++)
      panels->panels[i].outside=outside;
  }
}

/*------------------------------------------------------------------------*/
void getTARGETPANEL(void *panels_,const mxArray *rhs)
{
  panellist* panels = (panellist *)panels_;
  double outside=0.0;
  panels->slopecorrection=0;
  panels->aN=0;
  panels->bN=0;
  static const getgeneralpanelvalFun getval[] = {&getQUINTSTRENGTH,NULL,NULL,&getQUINTSHAPE,&getLINSTRENGTH,&getNORMALIZED,&getSHIFTED,
                                                 &getOUTSIDE,&getCUBSHAPE,&getPOLYSHAPE,&getPOLYSTRENGTH,NULL,&getSEPTSHAPE,
                                                 &getSEPTSTRENGTH,NULL,NULL,&getCUBSTRENGTH,&getLINSHAPE,NULL,&getSLOPECORRECTION,NULL};
  void *val[] = {panels,NULL,NULL,panels,panels,panels,panels,
                 &outside,panels,panels,panels,NULL,panels,
                 panels,NULL,NULL,panels,panels,NULL,panels,NULL};

  if(!mxIsCell(rhs))
    mexErrMsgTxt("targetpanel data must be given as cell object");
  int npanelrhs=mxGetNumberOfElements(rhs);
  // parse panel property/value-combinations
  for (int j = 0; j < npanelrhs-1; j++) {
    int prop;
    char str[17];
    const mxArray *Prop = mxGetCell(rhs,j);

    // determine property
    if (!mxIsChar(Prop))
      mexErrMsgTxt("Expecting a panel property name.");
    if (mxGetString(Prop,str,16) != 0 ||
      (prop = getgeneralpanelprop(str)) == ERR_PROPERTY)
      mexErrMsgTxt("Unknown panel property.");

    if(prop==POLYSTRENGTH||prop==LINSTRENGTH||prop==CUBSTRENGTH)
      mexErrMsgTxt("Panels strength should only be given to source panels");
    // get its value

    j+=getval[prop](val[prop],rhs,j+1,npanelrhs);
  }

  //finalize
  if(panels->aN==0)
    mexErrMsgTxt("Panel shape must be given");
  if(outside!=0.0) {
    for(int i=0;i<panels->count;i++)
      panels->panels[i].outside=outside;
  }
  int bN=panels->aN-1;
  panels->bN=bN;

  for(int i=0;i<panels->count;i++) {
    initgeneralvarsb(panels->panels+i,bN);
    initbvectortarget(panels->panels+i,i);
  }
}
/*------------------------------------------------------------------------*/
int getgeneralpanelprop(const char *str)
/* Small perfect hash for the supported panel properties. */
{
  static const char *tab[] = {"quintstrength","","","quintshape","linstrength","normalized","shifted",
                              "outside","cubshape","polyshape","polystrength","","septshape",
                              "septstrength","","","cubstrength","linshape","","slopecorrection",""};
  const int hash = ((int)str[0]+((int)str[6] * 5)) % 21;

  return str[0] != '\0' && strcmp(str,tab[hash]) == 0 ? 
    (PANELPROPERTY)hash : ERR_PROPERTY;
}
/*------------------------------------------------------------------------*/
int getSLOPECORRECTION(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  mxArray* arr=mxGetCell(rhs,j);

  if(panels->bN!=0)
    mexErrMsgTxt("slope correction must be given before panel strength");

  if(ISREALSCALAR(arr))
    panels->slopecorrection=(int)mxGetScalar(arr);
  else if (ISLOGICALSCALAR(arr))
    panels->slopecorrection = mxIsLogicalScalarTrue(arr);
  else
    mexErrMsgTxt("Expecting a logical scalar option 'slopecorrection'.");

  return 1;
}
/*------------------------------------------------------------------------*/
int getNORMALIZED(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  mxArray* arr=mxGetCell(rhs,j);

  if(ISREALSCALAR(arr))
    panels->normalized=(int)mxGetScalar(arr);
  else if (ISLOGICALSCALAR(arr))
    panels->normalized = mxIsLogicalScalarTrue(arr);
  else
    mexErrMsgTxt("Expecting a logical scalar option 'normalized'.");

  return 1;
}
/*------------------------------------------------------------------------*/
int getSHIFTED(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  mxArray* arr=mxGetCell(rhs,j);

  if(ISREALSCALAR(arr))
    panels->shifted=(int)mxGetScalar(arr);
  else if (ISLOGICALSCALAR(arr))
    panels->shifted = mxIsLogicalScalarTrue(arr);
  else
    mexErrMsgTxt("Expecting a logical scalar option 'shifted'.");

  return 1;
}
/*------------------------------------------------------------------------*/
int getOUTSIDE(void *outside_,const mxArray *rhs,int j,int maxsize)
{
  double *outside=(double*)outside_;
  mxArray* arr=mxGetCell(rhs,j);
  if (!ISREALSCALAR(arr))
      mexErrMsgTxt("Expecting a real and scalar option 'outside'.");
  *outside = *mxGetPr(arr);
  return 1;
}
/*------------------------------------------------------------------------*/
int getPOLYSHAPE(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->aN!=0)
    mexErrMsgTxt("panel shape already given");
  if(j>=maxsize-1)
    mexErrMsgTxt("polyshape must be followed by atleast two input arguments");
  mxArray* positions=mxGetCell(rhs,j);
  mxArray* a=mxGetCell(rhs,j+1);
  if(!ISDOUBLEMATRIX(positions))
    mexErrMsgTxt("First polyshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(a))
    mexErrMsgTxt("Second polyshape argument must be double matrix");
  if(mxGetN(positions)!=mxGetN(a))
    mexErrMsgTxt("both polyshape arguments must have same length");
  if(mxGetM(positions)!=2)
    mexErrMsgTxt("First polyshape argument must have 2 rows");

  double *xpositions=mxGetPr(positions);
  double *ypositions=mxGetPi(positions);
  double *ar=mxGetPr(a);
  double *ai=mxGetPi(a);
  int aN=mxGetM(a); //number of coefficients for panel shape
  int panelcount=mxGetN(positions);
  panels->aN=aN;
  panels->panels=(generalpanel*)myCalloc(panelcount,sizeof(generalpanel));
  panels->count=panelcount;
  panels->allocated=panelcount;
  for(int i=0;i<panelcount;i++) {
    if(ypositions!=NULL)
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],ypositions[2*i],xpositions[2*i+1], ypositions[2*i+1],0.0);
    else
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],0.0,xpositions[2*i+1], 0.0,0.0);
    
    initavectorgeneral(panels->panels+i,ar+aN*i,ai==NULL?NULL:ai+aN*i,aN);
  }
  return 2;
}
/*------------------------------------------------------------------------*/
int getLINSHAPE(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->aN!=0)
    mexErrMsgTxt("panel shape already given");
  mxArray* positions=mxGetCell(rhs,j);
  if(!ISDOUBLEMATRIX(positions))
    mexErrMsgTxt("linshape argument must be double matrix");

  if(mxGetM(positions)!=2)
    mexErrMsgTxt("linshape argument must have 2 rows");

  double *xpositions=mxGetPr(positions);
  double *ypositions=mxGetPi(positions);
  int aN=2; //number of coefficients for panel shape
  int panelcount=mxGetN(positions);
  panels->aN=aN;
  panels->panels=(generalpanel*)myCalloc(panelcount,sizeof(generalpanel));
  panels->count=panelcount;
  panels->allocated=panelcount;
  for(int i=0;i<panelcount;i++) {
    if(ypositions!=NULL)
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],ypositions[2*i],xpositions[2*i+1], ypositions[2*i+1],0.0);
    else
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],0.0,xpositions[2*i+1], 0.0,0.0);

    initavectorlinear(panels->panels+i);
  }
  return 1;
}
/*------------------------------------------------------------------------*/
int getCUBSHAPE(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->aN!=0)
    mexErrMsgTxt("panel shape already given");
  if(j>=maxsize-1)
    mexErrMsgTxt("cubshape must be followed by atleast two input arguments");
  mxArray* positions=mxGetCell(rhs,j);
  mxArray* directions=mxGetCell(rhs,j+1);
  if(!ISDOUBLEMATRIX(positions))
    mexErrMsgTxt("First cubshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(directions))
    mexErrMsgTxt("Second cubshape argument must be double matrix");
  if(mxGetN(positions)!=mxGetN(directions))
    mexErrMsgTxt("both cubshape arguments must have same length");
  if(mxGetM(positions)!=2)
    mexErrMsgTxt("First cubshape argument must have 2 rows");
  if(mxGetM(directions)!=2)
    mexErrMsgTxt("Second cubshape argument must have 2 rows");

  double *xpositions=mxGetPr(positions);
  double *ypositions=mxGetPi(positions);
  double *xdirections=mxGetPr(directions);
  double *ydirections=mxGetPi(directions);
  int aN=4; //number of coefficients for panel shape
  int panelcount=mxGetN(positions);
  panels->aN=aN;
  panels->panels=(generalpanel*)myCalloc(panelcount,sizeof(generalpanel));
  panels->count=panelcount;
  panels->allocated=panelcount;
  for(int i=0;i<panelcount;i++) {
    if(ypositions!=NULL)
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],ypositions[2*i],xpositions[2*i+1], ypositions[2*i+1],0.0);
    else
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],0.0,xpositions[2*i+1], 0.0,0.0);

    if(ydirections!=NULL) 
      initavectorcubic(panels->panels+i,xdirections[2*i],ydirections[2*i],xdirections[2*i+1],ydirections[2*i+1]);
    else
      initavectorcubic(panels->panels+i,xdirections[2*i],0.0,xdirections[2*i+1],0.0);
  }
  return 2;
}
/*------------------------------------------------------------------------*/
int getQUINTSHAPE(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->aN!=0)
    mexErrMsgTxt("panel shape already given");
  if(j>=maxsize-2)
    mexErrMsgTxt("quintshape must be followed by atleast three input arguments");
  mxArray* positions=mxGetCell(rhs,j);
  mxArray* directions=mxGetCell(rhs,j+1);
  mxArray* directions2=mxGetCell(rhs,j+2);
  if(!ISDOUBLEMATRIX(positions))
    mexErrMsgTxt("First quintshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(directions))
    mexErrMsgTxt("Second quintshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(directions2))
    mexErrMsgTxt("Third quintshape argument must be double matrix");
  if(mxGetN(positions)!=mxGetN(directions)||mxGetN(positions)!=mxGetN(directions2))
    mexErrMsgTxt("All three quintshape arguments must have same length");
  if(mxGetM(positions)!=2)
    mexErrMsgTxt("First quintshape argument must have 2 rows");
  if(mxGetM(directions)!=2)
    mexErrMsgTxt("Second quintshape argument must have 2 rows");
  if(mxGetM(directions2)!=2)
    mexErrMsgTxt("Third quintshape argument must have 2 rows");

  double *xpositions=mxGetPr(positions);
  double *ypositions=mxGetPi(positions);
  double *xdirections=mxGetPr(directions);
  double *ydirections=mxGetPi(directions);
  double *x2directions=mxGetPr(directions2);
  double *y2directions=mxGetPi(directions2);
  int aN=6; //number of coefficients for panel shape
  int panelcount=mxGetN(positions);
  panels->aN=aN;
  panels->panels=(generalpanel*)myCalloc(panelcount,sizeof(generalpanel));
  panels->count=panelcount;
  panels->allocated=panelcount;
  for(int i=0;i<panelcount;i++) {
    if(ypositions!=NULL)
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],ypositions[2*i],xpositions[2*i+1], ypositions[2*i+1],0.0);
    else
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],0.0,xpositions[2*i+1], 0.0,0.0);

    if(ydirections!=NULL) {
      if(y2directions!=NULL)
        initavectorquintic(panels->panels+i,xdirections[2*i],ydirections[2*i],xdirections[2*i+1],ydirections[2*i+1],x2directions[2*i],y2directions[2*i],x2directions[2*i+1],y2directions[2*i+1]);
      else
        initavectorquintic(panels->panels+i,xdirections[2*i],ydirections[2*i],xdirections[2*i+1],ydirections[2*i+1],x2directions[2*i],0.0,x2directions[2*i+1],0.0);
    }
    else {
      if(y2directions!=NULL)
        initavectorquintic(panels->panels+i,xdirections[2*i],0.0,xdirections[2*i+1],0.0,x2directions[2*i],y2directions[2*i],x2directions[2*i+1],y2directions[2*i+1]);
      else
        initavectorquintic(panels->panels+i,xdirections[2*i],0.0,xdirections[2*i+1],0.0,x2directions[2*i],0.0,x2directions[2*i+1],0.0);
    }
  }
  return 3;
}
/*------------------------------------------------------------------------*/
int getSEPTSHAPE(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->aN!=0)
    mexErrMsgTxt("panel shape already given");
  if(j>=maxsize-3)
    mexErrMsgTxt("septshape must be followed by atleast four input arguments");
  mxArray* positions=mxGetCell(rhs,j);
  mxArray* directions=mxGetCell(rhs,j+1);
  mxArray* directions2=mxGetCell(rhs,j+2);
  mxArray* directions3=mxGetCell(rhs,j+3);
  if(!ISDOUBLEMATRIX(positions))
    mexErrMsgTxt("First septshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(directions))
    mexErrMsgTxt("Second septshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(directions2))
    mexErrMsgTxt("Third septshape argument must be double matrix");
  if(!ISDOUBLEMATRIX(directions3))
    mexErrMsgTxt("Fourth septshape argument must be double matrix");
  if(mxGetN(positions)!=mxGetN(directions)||mxGetN(positions)!=mxGetN(directions2)||mxGetN(positions)!=mxGetN(directions3))
    mexErrMsgTxt("All four septshape arguments must have same length");
  if(mxGetM(positions)!=2)
    mexErrMsgTxt("First septshape argument must have 2 rows");
  if(mxGetM(directions)!=2)
    mexErrMsgTxt("Second septshape argument must have 2 rows");
  if(mxGetM(directions2)!=2)
    mexErrMsgTxt("Third septshape argument must have 2 rows");
  if(mxGetM(directions3)!=2)
    mexErrMsgTxt("Fourth septshape argument must have 2 rows");

  double *xpositions=mxGetPr(positions);
  double *ypositions=mxGetPi(positions);
  double *xdirections=mxGetPr(directions);
  double *ydirections=mxGetPi(directions);
  double *x2directions=mxGetPr(directions2);
  double *y2directions=mxGetPi(directions2);
  double *x3directions=mxGetPr(directions3);
  double *y3directions=mxGetPi(directions3);
  int aN=8; //number of coefficients for panel shape
  int panelcount=mxGetN(positions);
  panels->aN=aN;
  panels->panels=(generalpanel*)myCalloc(panelcount,sizeof(generalpanel));
  panels->count=panelcount;
  panels->allocated=panelcount;
  for(int i=0;i<panelcount;i++) {
    if(ypositions!=NULL)
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],ypositions[2*i],xpositions[2*i+1], ypositions[2*i+1],0.0);
    else
      initgeneralvarsa(panels->panels+i,aN,xpositions[2*i],0.0,xpositions[2*i+1], 0.0,0.0);
    double ydir1=0.0,ydir2=0.0;
    if(ydirections!=NULL) {
      ydir1=ydirections[2*i];
      ydir2=ydirections[2*i+1];
    }
    double y2dir1=0.0,y2dir2=0.0;
    if(y2directions!=NULL) {
      y2dir1=y2directions[2*i];
      y2dir2=y2directions[2*i+1];
    }
    double y3dir1=0.0,y3dir2=0.0;
    if(y3directions!=NULL) {
      y3dir1=y3directions[2*i];
      y3dir2=y3directions[2*i+1];
    }
    initavectorseptic(panels->panels+i,xdirections[2*i],ydir1,xdirections[2*i+1],ydir2,x2directions[2*i],y2dir1,x2directions[2*i+1],y2dir2,x3directions[2*i],y3dir1,x3directions[2*i+1],y3dir2);
  }
  return 4;
}
/*------------------------------------------------------------------------*/
int getPOLYSTRENGTH(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->bN!=0)
    mexErrMsgTxt("panel strength already given");

  mxArray* b=mxGetCell(rhs,j);
  if(!ISDOUBLEMATRIX(b))
    mexErrMsgTxt("polystrength argument must be double matrix");

  if(panels->aN==0)
    mexErrMsgTxt("polystregth must be given after shape information");
  if(panels->count!=mxGetN(b))
    mexErrMsgTxt("polystrength must have same length as shape");

  double *br=mxGetPr(b);
  double *bi=mxGetPi(b);
  int bN=mxGetM(b); //number of coefficients for panel shape
  panels->bN=bN;

  for(int i=0;i<panels->count;i++) {
    initgeneralvarsb(panels->panels+i,bN);
    initbvectorgeneral(panels->panels+i,br+i*bN,bi==NULL?NULL:bi+i*bN,bN);
  }
  return 1;
}
/*------------------------------------------------------------------------*/
int getLINSTRENGTH(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->bN!=0)
    mexErrMsgTxt("panel strength already given");

  mxArray* str=mxGetCell(rhs,j);
  if(!ISDOUBLEMATRIX(str))
    mexErrMsgTxt("linstrength argument must be double matrix");

  if(panels->aN==0)
    mexErrMsgTxt("linstregth must be given after shape information");
  if(panels->count!=mxGetN(str))
    mexErrMsgTxt("linstrength must have same length as shape");
  if(mxGetM(str)!=2)
    mexErrMsgTxt("linstrength argument must have 2 rows");

  double *strengths=mxGetPr(str);
  double *istrengths=mxGetPi(str);
  int bN=2;
  panels->bN=bN;

  for(int i=0;i<panels->count;i++) {
    initgeneralvarsb(panels->panels+i,bN);
    if(istrengths!=NULL)
      initbvectorlinear(panels->panels+i,strengths[2*i],istrengths[2*i],strengths[2*i+1],istrengths[2*i+1],panels->slopecorrection);
    else
      initbvectorlinear(panels->panels+i,strengths[2*i],0.0,strengths[2*i+1],0.0,panels->slopecorrection);
  }
  return 1;
}
/*------------------------------------------------------------------------*/
int getCUBSTRENGTH(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->bN!=0)
    mexErrMsgTxt("panel strength already given");

  if(j>=maxsize-1)
    mexErrMsgTxt("cubstrength must be followed by atleast two input arguments");
  mxArray* str=mxGetCell(rhs,j);
  mxArray* strdirections=mxGetCell(rhs,j+1);
  if(!ISDOUBLEMATRIX(str))
    mexErrMsgTxt("First cubstrength argument must be double matrix");
  if(!ISDOUBLEMATRIX(strdirections))
    mexErrMsgTxt("Second cubstrength argument must be double matrix");
  if(mxGetN(str)!=mxGetN(strdirections))
    mexErrMsgTxt("Both cubstrength arguments must have same length");
  if(mxGetM(str)!=2)
    mexErrMsgTxt("First cubstrength argument must have 2 rows");
  if(mxGetM(strdirections)!=2)
    mexErrMsgTxt("Second cubstrength argument must have 2 rows");

  if(panels->aN==0)
    mexErrMsgTxt("cubstregth must be given after shape information");
  if(panels->count!=mxGetN(str))
    mexErrMsgTxt("cubstrength must have same length as shape");

  double *strengths=mxGetPr(str);
  double *istrengths=mxGetPi(str);
  double *strengthdirections=mxGetPr(strdirections);
  double *istrengthdirections=mxGetPi(strdirections);
  int bN=4;
  panels->bN=bN;

  for(int i=0;i<panels->count;i++) {
    initgeneralvarsb(panels->panels+i,bN);
    if(istrengths!=NULL) {
      if(istrengthdirections!=NULL)
        initbvectorcubic(panels->panels+i,strengths[2*i],istrengths[2*i],strengths[2*i+1],istrengths[2*i+1],strengthdirections[2*i],istrengthdirections[2*i],strengthdirections[2*i+1],istrengthdirections[2*i+1],panels->slopecorrection);
      else
        initbvectorcubic(panels->panels+i,strengths[2*i],istrengths[2*i],strengths[2*i+1],istrengths[2*i+1],strengthdirections[2*i],0.0,strengthdirections[2*i+1],0.0,panels->slopecorrection);
    }
    else {
      if(istrengthdirections!=NULL)
        initbvectorcubic(panels->panels+i,strengths[2*i],0.0,strengths[2*i+1],0.0,strengthdirections[2*i],istrengthdirections[2*i],strengthdirections[2*i+1],istrengthdirections[2*i+1],panels->slopecorrection);
      else
        initbvectorcubic(panels->panels+i,strengths[2*i],0.0,strengths[2*i+1],0.0,strengthdirections[2*i],0.0,strengthdirections[2*i+1],0.0,panels->slopecorrection);
    }
  }
  return 2;
}
/*------------------------------------------------------------------------*/
int getQUINTSTRENGTH(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->bN!=0)
    mexErrMsgTxt("panel strength already given");

  if(j>=maxsize-2)
    mexErrMsgTxt("quintstrength must be followed by atleast three input arguments");
  mxArray* str=mxGetCell(rhs,j);
  mxArray* strdirections=mxGetCell(rhs,j+1);
  mxArray* str2directions=mxGetCell(rhs,j+2);
  if(!ISDOUBLEMATRIX(str))
    mexErrMsgTxt("First quintstrength argument must be double matrix");
  if(!ISDOUBLEMATRIX(strdirections))
    mexErrMsgTxt("Second quintstrength argument must be double matrix");
  if(!ISDOUBLEMATRIX(str2directions))
    mexErrMsgTxt("Third quintstrength argument must be double matrix");
  if(mxGetN(str)!=mxGetN(strdirections)||mxGetN(str)!=mxGetN(str2directions))
    mexErrMsgTxt("All quintstrength arguments must have same length");
  if(mxGetM(str)!=2)
    mexErrMsgTxt("First quintstrength argument must have 2 rows");
  if(mxGetM(strdirections)!=2)
    mexErrMsgTxt("Second quintstrength argument must have 2 rows");
  if(mxGetM(str2directions)!=2)
    mexErrMsgTxt("Third quintstrength argument must have 2 rows");

  if(panels->aN==0)
    mexErrMsgTxt("quintstregth must be given after shape information");
  if(panels->count!=mxGetN(str))
    mexErrMsgTxt("quintstregth must have same length as shape");

  double *strengths=mxGetPr(str);
  double *istrengths=mxGetPi(str);
  double *strengthdirections=mxGetPr(strdirections);
  double *istrengthdirections=mxGetPi(strdirections);
  double *strength2directions=mxGetPr(str2directions);
  double *istrength2directions=mxGetPi(str2directions);

  int bN=6;
  panels->bN=bN;

  for(int i=0;i<panels->count;i++) {
    initgeneralvarsb(panels->panels+i,bN);
    double istrstart=0;
    double istrend=0;
    if(istrengths!=NULL) {
      istrstart=istrengths[2*i];
      istrend=istrengths[2*i+1];
    }
    double istrdirstart=0;
    double istrdirend=0;
    if(istrengthdirections!=NULL) {
      istrdirstart=istrengthdirections[2*i];
      istrdirend=istrengthdirections[2*i+1];
    }
    double istr2dirstart=0;
    double istr2dirend=0;
    if(istrength2directions!=NULL) {
      istr2dirstart=istrength2directions[2*i];
      istr2dirend=istrength2directions[2*i+1];
    }
    initbvectorquintic(panels->panels+i,strengths[2*i],istrstart,strengths[2*i+1],istrend,strengthdirections[2*i],istrdirstart,strengthdirections[2*i+1],istrdirend,strength2directions[2*i],istr2dirstart,strength2directions[2*i+1],istr2dirend,panels->slopecorrection);
  }
  return 3;
}
/*------------------------------------------------------------------------*/
int getSEPTSTRENGTH(void *panels_,const mxArray *rhs,int j,int maxsize)
{
  panellist* panels=(panellist*)panels_;
  if(panels->bN!=0)
    mexErrMsgTxt("panel strength already given");

  if(j>=maxsize-3)
    mexErrMsgTxt("septstrength must be followed by atleast four input arguments");
  mxArray* str=mxGetCell(rhs,j);
  mxArray* strdirections=mxGetCell(rhs,j+1);
  mxArray* str2directions=mxGetCell(rhs,j+2);
  mxArray* str3directions=mxGetCell(rhs,j+3);
  if(!ISDOUBLEMATRIX(str))
    mexErrMsgTxt("First septstrength argument must be double matrix");
  if(!ISDOUBLEMATRIX(strdirections))
    mexErrMsgTxt("Second septstrength argument must be double matrix");
  if(!ISDOUBLEMATRIX(str2directions))
    mexErrMsgTxt("Third septstrength argument must be double matrix");
  if(!ISDOUBLEMATRIX(str3directions))
    mexErrMsgTxt("Fourth septstrength argument must be double matrix");
  if(mxGetN(str)!=mxGetN(strdirections)||mxGetN(str)!=mxGetN(str2directions))
    mexErrMsgTxt("All septstrength arguments must have same length");
  if(mxGetM(str)!=2)
    mexErrMsgTxt("First septstrength argument must have 2 rows");
  if(mxGetM(strdirections)!=2)
    mexErrMsgTxt("Second septstrength argument must have 2 rows");
  if(mxGetM(str2directions)!=2)
    mexErrMsgTxt("Third septstrength argument must have 2 rows");
  if(mxGetM(str3directions)!=2)
    mexErrMsgTxt("Fourth septstrength argument must have 2 rows");

  if(panels->aN==0)
    mexErrMsgTxt("septstrength must be given after shape information");
  if(panels->count!=mxGetN(str))
    mexErrMsgTxt("septstrength must have same length as shape");

  double *strengths=mxGetPr(str);
  double *istrengths=mxGetPi(str);
  double *strengthdirections=mxGetPr(strdirections);
  double *istrengthdirections=mxGetPi(strdirections);
  double *strength2directions=mxGetPr(str2directions);
  double *istrength2directions=mxGetPi(str2directions);
  double *strength3directions=mxGetPr(str3directions);
  double *istrength3directions=mxGetPi(str3directions);

  int bN=8;
  panels->bN=bN;

  for(int i=0;i<panels->count;i++) {
    initgeneralvarsb(panels->panels+i,bN);
    double istrstart=0;
    double istrend=0;
    if(istrengths!=NULL) {
      istrstart=istrengths[2*i];
      istrend=istrengths[2*i+1];
    }
    double istrdirstart=0;
    double istrdirend=0;
    if(istrengthdirections!=NULL) {
      istrdirstart=istrengthdirections[2*i];
      istrdirend=istrengthdirections[2*i+1];
    }
    double istr2dirstart=0;
    double istr2dirend=0;
    if(istrength2directions!=NULL) {
      istr2dirstart=istrength2directions[2*i];
      istr2dirend=istrength2directions[2*i+1];
    }
    double istr3dirstart=0;
    double istr3dirend=0;
    if(istrength3directions!=NULL) {
      istr3dirstart=istrength3directions[2*i];
      istr3dirend=istrength3directions[2*i+1];
    }
    initbvectorseptic(panels->panels+i,strengths[2*i],istrstart,strengths[2*i+1],istrend,strengthdirections[2*i],istrdirstart,strengthdirections[2*i+1],istrdirend,strength2directions[2*i],istr2dirstart,strength2directions[2*i+1],istr2dirend,strength3directions[2*i],istr3dirstart,strength3directions[2*i+1],istr3dirend,panels->slopecorrection);
  }
  return 4;
}
/*------------------------------------------------------------------------*/

