//input1: panel positions, two row vector with start and end points
//input2: a-coefficients for the panel shape, real valued, should give y-position when panel is aligned between start and ens points a[1]*x^1+a[2]*x^2+a[3]*x^3 etc. a[0] will not be used but should be included
//input3: b-coefficients for the vorticity shape, possibly complex valued, should give vorticity as function of when panel is aligned between start and ens points b[0]+b[1]*x^1+b[2]*x^2+b[3]*x^3
//input4: evaluation points
//input5: outside parameter
#include <math.h>
#include <float.h>
#include "mex.h"
#include "matrix.h"
#include "generalpanel.h"
#ifdef _MSC_VER
#define copysign _copysign
#define hypot _hypot
#endif


#define ISDOUBLEMATRIX(A) (mxIsDouble(A) && !mxIsSparse(A) && \
			   mxGetNumberOfDimensions(A) == 2)
#define ISREALSCALAR(A) (!mxIsCell(A) && !mxIsStruct(A) && \
			 !mxIsComplex(A) && mxGetNumberOfElements(A) == 1)
 
void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
  const int Ncoeff=GENERALPANELNCOEFF;
  double *xvelocities,*yvelocities;
  double *vortexxpoint,*vortexypoint;
  double outside;
  int slopecorrection=1;
  if (nrhs != 3 && nrhs != 4)
    mexErrMsgTxt("Expecting 3 or 4 inputs.");

  if (!ISDOUBLEMATRIX(prhs[0]) || !ISDOUBLEMATRIX(prhs[1]))
    mexErrMsgTxt("First two arguments must be double matrices.");

//   if(!mxIsComplex(prhs[0]) || !mxIsComplex(prhs[1]) || !mxIsComplex(prhs[2]))
//       mexErrMsgTxt("First three arguments must be complex.");
  
  if(mxGetN(prhs[1])!=mxGetN(prhs[2])||nrhs==4&&mxGetN(prhs[1])!=mxGetN(prhs[3]))
    mexErrMsgTxt("Arguments 2-4 must have same length");
  if(mxGetM(prhs[1])!=2)
    mexErrMsgTxt("Second argument must have 2 rows");

  
  double *splitpoints=mxGetPr(prhs[0]);
  double *xpositions=mxGetPr(prhs[1]);
  double *ypositions=mxGetPi(prhs[1]);
  double *ar=mxGetPr(prhs[2]);
  double *ai=mxGetPi(prhs[2]);
  double *br=NULL;
  double *bi=NULL;
  
  double *outputar=NULL;
  double *outputai=NULL;
  double *outputbr=NULL;
  double *outputbi=NULL;
  int bN=1;
  
  if(nrhs==4) {
    br=mxGetPr(prhs[3]);
    bi=mxGetPi(prhs[3]);
	bN=mxGetM(prhs[3]);
  }
  
  int splitpointlength=mxGetNumberOfElements(prhs[0]);
  int aN=mxGetM(prhs[2]); //number of coefficients for panel shape
  

  int panelcount=mxGetN(prhs[1]);
  plhs[0] = mxCreateDoubleMatrix(2,(splitpointlength+1)*panelcount,mxCOMPLEX);
  double *xoutputpositions=mxGetPr(plhs[0]);
  double *youtputpositions=mxGetPi(plhs[0]);
  if(nlhs>1) {
    if(mxIsComplex(prhs[2]))
      plhs[1]=mxCreateDoubleMatrix(aN,(splitpointlength+1)*panelcount,mxCOMPLEX);
    else
      plhs[1]=mxCreateDoubleMatrix(aN,(splitpointlength+1)*panelcount,mxREAL);
    outputar=mxGetPr(plhs[1]);
    outputai=mxGetPi(plhs[1]);
  }
  if(nlhs>2) {
    if(mxIsComplex(prhs[3]))
      plhs[2]=mxCreateDoubleMatrix(bN,(splitpointlength+1)*panelcount,mxCOMPLEX);
    else
      plhs[2]=mxCreateDoubleMatrix(bN,(splitpointlength+1)*panelcount,mxREAL);
    outputbr=mxGetPr(plhs[2]);
    outputbi=mxGetPi(plhs[2]);
  }
  
  generalpanelalloc(aN,bN,Ncoeff,1);
  panellist panels;
  panels.panels=(generalpanel*)myCalloc(panelcount*(splitpointlength+1),sizeof(generalpanel));
  panels.count=panelcount*(splitpointlength+1);
  panels.allocated=panelcount*(splitpointlength+1);
  for(int i=0;i<panelcount;i++) {
    if(ypositions!=NULL)
      initgeneralpanel(panels.panels+i*(splitpointlength+1),xpositions[2*i],ypositions[2*i],xpositions[2*i+1], ypositions[2*i+1],ar+aN*i,ai==NULL?NULL:(ai+aN*i),aN,br==NULL?br:(br+bN*i),bi==NULL?bi:(bi+bN*i),bN,outside);
    else
      initgeneralpanel(panels.panels+i*(splitpointlength+1),xpositions[2*i],0.0,xpositions[2*i+1], 0.0,ar+aN*i,ai==NULL?NULL:(ai+aN*i),aN,br==NULL?br:(br+bN*i),bi==NULL?bi:(bi+bN*i),bN,outside);
    initgeneralpanelcvector(panels.panels+i*(splitpointlength+1));
    double endval=panels.panels[i*(splitpointlength+1)].endval;
    for(int j=splitpointlength-1;j>=0;j--)
      performsplit(i*(splitpointlength+1),i*(splitpointlength+1), &panels, splitpoints[j]*endval, panels.panels[i*(splitpointlength+1)].endval,0,0,i*(splitpointlength+1)+j+1);

    //initgeneralpanelconnectingterms(panels.panels+i);
    //initgeneralpanelbox(panels.panels+i);
    //initgeneralpaneldist(panels.panels+i);
    //initgeneralpanelexpansion(panels.panels+i,Ncoeff);
//    printgeneralpanel(panels+i);
    /*if(panels.panels[i].a_order==3) {
      if(panels.panels[i].areal)
        initcubicconstants_areal(&panels.panels[i].cub,panels.panels+i);
      else
        initcubicconstants_aimag(&panels.panels[i].cub,panels.panels+i);
    }*/
    //printgeneralpanel(panels.panels+i);
//    generalpaneldirectinteract(xvelocities,yvelocities,panels+i,evalxpoint,evalypoint,evalxcount*evalycount,wksp1big,wksp1,wksp2,wksp2big,wksp3,wkspi);
//    freegeneralpanel(panels+i);
  }
  for(int i=0;i<panels.count;i++) {
    xoutputpositions[2*i]=panels.panels[i].z1.re;
    xoutputpositions[2*i+1]=panels.panels[i].z2.re;
    youtputpositions[2*i]=panels.panels[i].z1.im;
    youtputpositions[2*i+1]=panels.panels[i].z2.im;
    if(nlhs>1) {
      for(int j=0;j<=panels.panels[i].a_order;j++) {
        outputar[aN*i+j]=panels.panels[i].a[j].re;
        if(outputai!=NULL)
          outputai[aN*i+j]=panels.panels[i].a[j].im;
      }
    }
    if(nlhs>2) {
      for(int j=0;j<=panels.panels[i].a_order;j++) {
        outputar[aN*i+j]=panels.panels[i].a[j].re;
        if(outputai!=NULL)
          outputai[aN*i+j]=panels.panels[i].a[j].im;
      }
    }
    freegeneralpanel(panels.panels+i);
  }
  generalpanelfree();
}
